@@ -23,15 +23,22 @@ using Rcpp::stop;
 using Rcpp::wrap;
 
 #if defined(ARROW_R_WITH_ARROW)
+#include <arrow/type.h>
+
+RCPP_EXPOSED_ENUM_NODECL(arrow::Type::type)
+RCPP_EXPOSED_ENUM_NODECL(arrow::DateUnit)
+RCPP_EXPOSED_ENUM_NODECL(arrow::TimeUnit::type)
 
 // [[arrow::export]]
 bool shared_ptr_is_null(SEXP xp) {
-  return reinterpret_cast<std::shared_ptr<void>*>(EXTPTR_PTR(xp))->get() == nullptr;
+  return reinterpret_cast<std::shared_ptr<void>*>(R_ExternalPtrAddr(xp))->get() ==
+         nullptr;
 }
 
 // [[arrow::export]]
 bool unique_ptr_is_null(SEXP xp) {
-  return reinterpret_cast<std::unique_ptr<void>*>(EXTPTR_PTR(xp))->get() == nullptr;
+  return reinterpret_cast<std::unique_ptr<void>*>(R_ExternalPtrAddr(xp))->get() ==
+         nullptr;
 }
 
 // [[arrow::export]]
@@ -73,9 +80,17 @@ std::shared_ptr<arrow::DataType> Boolean__initialize() { return arrow::boolean()
 // [[arrow::export]]
 std::shared_ptr<arrow::DataType> Utf8__initialize() { return arrow::utf8(); }
 
+// [[arrow::export]]
+std::shared_ptr<arrow::DataType> LargeUtf8__initialize() { return arrow::large_utf8(); }
+
 // [[arrow::export]]
 std::shared_ptr<arrow::DataType> Binary__initialize() { return arrow::binary(); }
 
+// [[arrow::export]]
+std::shared_ptr<arrow::DataType> LargeBinary__initialize() {
+  return arrow::large_binary();
+}
+
 // [[arrow::export]]
 std::shared_ptr<arrow::DataType> Date32__initialize() { return arrow::date32(); }
 
@@ -89,7 +104,7 @@ std::shared_ptr<arrow::DataType> Null__initialize() { return arrow::null(); }
 std::shared_ptr<arrow::DataType> Decimal128Type__initialize(int32_t precision,
                                                             int32_t scale) {
   // Use the builder that validates inputs
-  return VALUE_OR_STOP(arrow::Decimal128Type::Make(precision, scale));
+  return ValueOrStop(arrow::Decimal128Type::Make(precision, scale));
 }
 
 // [[arrow::export]]
@@ -135,6 +150,38 @@ SEXP list__(SEXP x) {
   return R_NilValue;
 }
 
+// [[arrow::export]]
+SEXP large_list__(SEXP x) {
+  if (Rf_inherits(x, "Field")) {
+    Rcpp::ConstReferenceSmartPtrInputParameter<std::shared_ptr<arrow::Field>> field(x);
+    return wrap(arrow::large_list(field));
+  }
+
+  if (Rf_inherits(x, "DataType")) {
+    Rcpp::ConstReferenceSmartPtrInputParameter<std::shared_ptr<arrow::DataType>> type(x);
+    return wrap(arrow::large_list(type));
+  }
+
+  stop("incompatible");
+  return R_NilValue;
+}
+
+// [[arrow::export]]
+SEXP fixed_size_list__(SEXP x, int list_size) {
+  if (Rf_inherits(x, "Field")) {
+    Rcpp::ConstReferenceSmartPtrInputParameter<std::shared_ptr<arrow::Field>> field(x);
+    return wrap(arrow::fixed_size_list(field, list_size));
+  }
+
+  if (Rf_inherits(x, "DataType")) {
+    Rcpp::ConstReferenceSmartPtrInputParameter<std::shared_ptr<arrow::DataType>> type(x);
+    return wrap(arrow::fixed_size_list(type, list_size));
+  }
+
+  stop("incompatible");
+  return R_NilValue;
+}
+
 // [[arrow::export]]
 std::shared_ptr<arrow::DataType> struct_(List fields) {
   return arrow::struct_(arrow::r::List_to_shared_ptr_vector<arrow::Field>(fields));
@@ -158,12 +205,12 @@ bool DataType__Equals(const std::shared_ptr<arrow::DataType>& lhs,
 
 // [[arrow::export]]
 int DataType__num_children(const std::shared_ptr<arrow::DataType>& type) {
-  return type->num_children();
+  return type->num_fields();
 }
 
 // [[arrow::export]]
 List DataType__children_pointer(const std::shared_ptr<arrow::DataType>& type) {
-  return List(type->children().begin(), type->children().end());
+  return List(type->fields().begin(), type->fields().end());
 }
 
 // [[arrow::export]]
@@ -216,7 +263,7 @@ arrow::TimeUnit::type TimestampType__unit(
 std::shared_ptr<arrow::DataType> DictionaryType__initialize(
     const std::shared_ptr<arrow::DataType>& index_type,
     const std::shared_ptr<arrow::DataType>& value_type, bool ordered) {
-  return VALUE_OR_STOP(arrow::DictionaryType::Make(index_type, value_type, ordered));
+  return ValueOrStop(arrow::DictionaryType::Make(index_type, value_type, ordered));
 }
 
 // [[arrow::export]]
@@ -265,4 +312,33 @@ std::shared_ptr<arrow::DataType> ListType__value_type(
   return type->value_type();
 }
 
+// [[arrow::export]]
+std::shared_ptr<arrow::Field> LargeListType__value_field(
+    const std::shared_ptr<arrow::LargeListType>& type) {
+  return type->value_field();
+}
+
+// [[arrow::export]]
+std::shared_ptr<arrow::DataType> LargeListType__value_type(
+    const std::shared_ptr<arrow::LargeListType>& type) {
+  return type->value_type();
+}
+
+// [[arrow::export]]
+std::shared_ptr<arrow::Field> FixedSizeListType__value_field(
+    const std::shared_ptr<arrow::FixedSizeListType>& type) {
+  return type->value_field();
+}
+
+// [[arrow::export]]
+std::shared_ptr<arrow::DataType> FixedSizeListType__value_type(
+    const std::shared_ptr<arrow::FixedSizeListType>& type) {
+  return type->value_type();
+}
+
+// [[arrow::export]]
+int FixedSizeListType__list_size(const std::shared_ptr<arrow::FixedSizeListType>& type) {
+  return type->list_size();
+}
+
 #endif
